#include "Theatre.h"

Theatre::Theatre(QWidget *parent)
    : QWidget(parent)
{
	// Default UI setup -- initializes all Qt objects
	ui.setupUi(this);

	// Customize layout
	configure();
	layout();

	// Initialize timer with 1s callback
	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(callback()));
	timer->start(1000);
	first_timeout = true;

	// Gstreamer initialization
	gst_player = new GstPlayer();
	gst_bus = NULL;

	// Gstreamer overrides alpha blend, so this sets it back to desired blend
	set_alpha(255);

	// Default: is not playing
	is_playing = false;
	is_stopped = true;

	// Initialize CPU control objects
	usage = new Usage(num_cores);
	cpu_control = new CPUControl(num_cores);

	// Reset the thread count to 0
	unload_cpu();

}

/**
 * Read configuration parameters from external file, and process to GUI
 * setup rules
 */
void
Theatre::configure()
{
	const int MAX_LENGTH = 200;
	char cores_buffer[MAX_LENGTH];
	char config_buffer[MAX_LENGTH];
	bool result;

	// Retrieve number of cores from configuration
	get_config("cores", cores_buffer);
	num_cores = atoi(cores_buffer);
	printf("Number of cores to monitor: %d\n", num_cores);

	if (get_config("pipeline", config_buffer) == false)
		ui.playNoAccelButton->setEnabled(false);

	if (get_config("pipeline_fast", config_buffer) == false)
		ui.playAccelButton->setEnabled(false);

	if (get_config("pipeline_camera", config_buffer) == false)
		ui.cameraButton->setEnabled(false);

}

/**
 * Initialize layout of the Qt GUI
 */
void
Theatre::layout()
{
	// Buffer between border and elements
	const int SPACE = 10;
	int pbw, pbh;

	QWidget::showFullScreen();

	// Set screen dimensions
	qdw = QApplication::desktop();
	QRect screen = qdw->availableGeometry(0);

	// Please don't hate for the hard-coded position numbers below...
	// this is Qt after all!

	// Transfer button width and height to shorter variable names
	pbw = ui.playPauseButton->width();
	pbh = ui.playPauseButton->height();

	// Move play button
	ui.playPauseButton->move(screen.right() - (pbw*2 + SPACE*7),
			screen.bottom() - (pbh + SPACE*4));

	// Move stop button
	ui.stopButton->move(screen.right() - (pbw + SPACE*5),
			screen.bottom() - (pbh + SPACE*4));

	// Extra 10s and 20s hide the dotted button border
	ui.fullButton->setGeometry(screen.left() - SPACE,
			screen.top() - SPACE, screen.width() + SPACE*2,
			screen.height() + SPACE*2);

	// Set Instruction Text
	ui.instructionLabel->move((screen.width() -
			ui.instructionLabel->width()) / 2, screen.height() / 2 -
			SPACE - ui.playAccelButton->height() -
			SPACE - ui.instructionLabel->height());

	// Set position of initial 'play' buttons
	ui.playAccelButton->move(screen.width() / 2 -
			ui.playAccelButton->width() -SPACE, screen.height() / 2 -
			SPACE - ui.playAccelButton->height());
	ui.playNoAccelButton->move(screen.width() / 2 + SPACE,
			screen.height() / 2 - SPACE-
			ui.playNoAccelButton->height());
	ui.cameraButton->move(screen.width() / 2 -
			ui.cameraButton->width() - SPACE,
			screen.height() / 2 + SPACE);
	ui.exitButton->move(screen.width() / 2 + SPACE,
			screen.height() / 2 +  SPACE);

	// Disable buttons based on the functionality configured


	// Position the label for CPU usage and hide it
	ui.usageLabel->move(SPACE*4, SPACE*4);

	// Reposition core-enable button
	ui.coreEnableButton->move(SPACE*5,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE -
			ui.coreEnableButton->height() - SPACE);

	ui.enableLabel->move(SPACE*5 + ui.coreEnableButton->width() + SPACE,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE -
			ui.enableLabel->height() - SPACE);

	// Reposition core label
	ui.coreSelectLabel->move(SPACE*5 + ui.coreSelectBox->width() + SPACE,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE -
			ui.enableLabel->height() - SPACE -
			ui.coreSelectLabel->height() - SPACE);

	// Reposition core select box
	ui.coreSelectBox->move(SPACE*5 ,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE -
			ui.enableLabel->height() - SPACE -
			ui.coreSelectBox->height() - SPACE);

	// Add entry for each non-zero CPU core
	for (unsigned int i = 1; i < num_cores; i++)
	{
		char buffer[1];
		sprintf(buffer, "%d", i);
		ui.coreSelectBox->addItem(QString::fromUtf8(buffer), NULL);
	}

	// Position copyright label
	ui.copyrightLabel->move((screen.width() - ui.copyrightLabel->width()) / 2,
			screen.height() - ui.copyrightLabel->height());

	// Position CPU buttons for adding workload
	ui.unloadCPUButton->move(SPACE*5, screen.height() -
			ui.unloadCPUButton->height() - SPACE*3);
	ui.unloadCPULabel->move(SPACE*5 + ui.unloadCPUButton->width() + SPACE,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3);

	ui.loadCPUButton->move(SPACE*5, screen.height() -
			ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE);
	ui.loadCPULabel->move(SPACE*5 + ui.loadCPUButton->width() + SPACE,
			screen.height() - ui.unloadCPUButton->height() - SPACE*3 -
			ui.loadCPUButton->height() - SPACE);

	// Add thread label
	ui.threadLabel->move(screen.width() - ui.threadLabel->width() - SPACE,
			SPACE*3);
	update_thread_count();

	// Hide items for the pause screen
	ui.coreEnableButton->hide();
	ui.coreSelectBox->hide();
	ui.coreSelectLabel->hide();
	ui.usageLabel->hide();
	ui.playPauseButton->hide();
	ui.stopButton->hide();
	ui.copyrightLabel->hide();
	ui.loadCPUButton->hide();
	ui.unloadCPUButton->hide();
	ui.enableLabel->hide();
	ui.loadCPULabel->hide();
	ui.unloadCPULabel->hide();
	ui.threadLabel->hide();

	// Re-order the Qt items in terms of which accept input first
	ui.fullButton->raise();
	ui.instructionLabel->raise();
	ui.usageLabel->raise();
	ui.playPauseButton->raise();
	ui.stopButton->raise();
	ui.coreEnableButton->raise();
	ui.coreSelectBox->raise();
	ui.coreSelectLabel->raise();
	ui.playAccelButton->raise();
	ui.playNoAccelButton->raise();
	ui.cameraButton->raise();
	ui.exitButton->raise();
	ui.copyrightLabel->raise();
	ui.loadCPUButton->raise();
	ui.unloadCPUButton->raise();
	ui.enableLabel->raise();
	ui.loadCPULabel->raise();
	ui.unloadCPULabel->raise();
	ui.threadLabel->raise();

	// Set up and hide the splash image
	splash_label = new QLabel(this);
	splash_pixmap = QPixmap(QString::fromUtf8(":/timesys.jpg"));
	splash_label->setPixmap(splash_pixmap);
	splash_label->resize(splash_pixmap.width(), splash_pixmap.height());
	splash_label->move((screen.width() - splash_label->width())/2,
			(screen.height() - splash_label->height() - SPACE*6));
	splash_label->hide();
	splash_palette.setBrush(this->backgroundRole(), QBrush(QColor(0, 0, 0)));
	this->setPalette(splash_palette);

	// Set cursor at center of the screen
	QCursor::setPos(screen.left() + screen.width()/2,
					screen.top() + screen.height()/2);
}

Theatre::~Theatre()
{
}

void
Theatre::clicked()
{
	if(is_playing)
	{
		set_alpha(127);
		ui.playPauseButton->show();
		ui.stopButton->show();

		ui.usageLabel->show();
		ui.coreEnableButton->show();
		ui.coreSelectBox->show();
		ui.coreSelectLabel->show();
		ui.copyrightLabel->show();
		ui.enableLabel->show();
		ui.loadCPULabel->show();
		ui.unloadCPULabel->show();
		ui.threadLabel->show();

		ui.loadCPUButton->show();
		ui.unloadCPUButton->show();
	}
}

void
Theatre::play_pause()
{
	// Play stream
	if(!is_playing){

		gst_player->play();
		ui.playPauseButton->setIcon(QIcon(":/pause.jpg"));

		paint_splash(false);
		ui.instructionLabel->hide();
		ui.playPauseButton->hide();
		ui.stopButton->hide();
		ui.playAccelButton->hide();
		ui.playNoAccelButton->hide();
		ui.cameraButton->hide();
		ui.exitButton->hide();
		ui.copyrightLabel->hide();
		ui.enableLabel->hide();
		ui.loadCPULabel->hide();
		ui.unloadCPULabel->hide();
		ui.threadLabel->hide();

		ui.usageLabel->hide();
		ui.coreEnableButton->hide();
		ui.coreSelectBox->hide();
		ui.coreSelectLabel->hide();
		ui.loadCPUButton->hide();
		ui.unloadCPUButton->hide();

		// If beginning to play the first time, we have to pause to let
		// GStreamer get its ducks in a row
		if(is_stopped)
		{
			usleep(1000000);
			is_stopped = false;
		}
		set_alpha(0);
	}
	// Pause stream
	else{
		gst_player->pause();
		ui.playPauseButton->setIcon(QIcon(":/play.jpg"));
	}

	is_playing = !is_playing; // Invert value of is_playing
}

// Reset pipeline to beginning of stream
void
Theatre::stop()
{
	reset_player();
	is_playing = false;
	is_stopped = true;
	ui.playPauseButton->setIcon(QIcon(":/play.jpg"));

	ui.instructionLabel->show();
	ui.playAccelButton->show();
	ui.playNoAccelButton->show();
	ui.cameraButton->show();
	ui.exitButton->show();

	ui.playPauseButton->hide();
	ui.stopButton->hide();
	ui.usageLabel->hide();
	ui.coreEnableButton->hide();
	ui.coreSelectBox->hide();
	ui.coreSelectLabel->hide();
	ui.copyrightLabel->hide();
	ui.loadCPUButton->hide();
	ui.unloadCPUButton->hide();
	ui.enableLabel->hide();
	ui.loadCPULabel->hide();
	ui.unloadCPULabel->hide();
	ui.threadLabel->hide();

	paint_splash(true);

	ui.playAccelButton->raise();
	ui.playNoAccelButton->raise();

	usleep(500000);
	set_alpha(255);
}

// Dereference pipeline, end program
void
Theatre::end()
{
	gst_player->destroy();

	delete(timer);
	delete(gst_player);

	exit(0);
}

void
Theatre::reset_player()
{
	// Only reset player if it has been fully initialized
	gst_player->stop();
	gst_bus = gst_player->get_bus();
	unload_cpu();
}

void
Theatre::callback()
{
	const char MAX_LENGTH = 200;
	char usage_string[MAX_LENGTH] = "\0";
	float core_usage;
	bool core_enabled;

	// Recompute the CPU usage all cores
	usage->recompute();

	// Append usage for each core to the display
	for (unsigned int i = 0; i < num_cores; i++)
	{
		core_enabled = cpu_control->get_enabled(i);
		core_usage = usage->get_usage(i);

		sprintf(usage_string, "%s CPU %d: %.f%%", usage_string, i, core_usage);

		if (core_enabled == true)
			sprintf(usage_string, "%s [On]", usage_string);
		else
			sprintf(usage_string, "%s [Off]", usage_string);

		sprintf(usage_string, "%s\n", usage_string);
	}

	// Reset usage message to new usage
	ui.usageLabel->setText(QString(usage_string));

	// If EOS reached, call bus
	if(check_bus())
	{
		stop();
	}
}

// Pass true if you want to fill with TimeSys logo/
// Otherwise, pass false for black
void
Theatre::paint_splash(bool fill)
{
	if(fill == true)
	{
		splash_palette.setBrush(this->backgroundRole(), QBrush(QColor(0, 84, 128)));
		this->setPalette(splash_palette);
		splash_label->show();
	}
	else
	{
		splash_palette.setBrush(this->backgroundRole(), QBrush(QColor(0, 0, 0)));
		this->setPalette(splash_palette);
		splash_label->hide();
	}
}

// Return true if EOS; else, returns false
bool
Theatre::check_bus()
{
	GstMessage *msg;
	bool eos_reached = false;

	// If this is a NULL pointer, just return
	if (gst_bus == NULL)
		return eos_reached;

	while((msg = gst_bus_pop(gst_bus)) != NULL){
		// g_print("\nBus got message!");
		switch(GST_MESSAGE_TYPE(msg)){
			case GST_MESSAGE_EOS: {
				g_print("\nEOS reached. Now stopping pipeline");
				eos_reached = true;
				break;
			}
			case GST_MESSAGE_ERROR: {
				g_print("\nGstreamer error");
				GError *error;
				gchar *debug;
				gst_message_parse_error(msg, &error, &debug);
				g_print("\nError: %s", error->message);
				g_error(debug);
				g_error_free(error);
				break;
			}
			default:
				break;
		}
		gst_message_unref(msg);
	}
	return eos_reached;
}

/**
 * Update the text on the "Enable/Disable" button
 */
void
Theatre::update_enabled_button()
{
	char MAX_LENGTH = 20;
	char button_text[MAX_LENGTH];
	int core_selected;
	bool core_enabled;

	// Find whether target core is enabled
	core_selected = ui.coreSelectBox->currentText().toInt(NULL, 10);
	core_enabled = cpu_control->get_enabled(core_selected);

	// Set button text to enable or disable CPU accordingly
	if (core_enabled == true)
		strcpy(button_text, "Disable core");
	else if (core_enabled == false)
		strcpy(button_text, "Enable core");

	ui.enableLabel->setText(QString::fromUtf8(button_text));
	printf("Button text is %s\n", button_text);

}

/**
 * Enable or disable a core after the "Enable/Disable" button is pressed
 */
void
Theatre::enable_button_clicked()
{
	QString action;
	int core_selected;
	bool enable_cpu = false;

	// Retrieve the index of the selected core
	core_selected = ui.coreSelectBox->currentText().toInt(NULL, 10);

	// Find whether the request is to enable or disable the core
	action = ui.enableLabel->text();
	if (action == "Disable core")
		enable_cpu = false;
	else if (action == "Enable core")
		enable_cpu = true;
	else
	{
		printf("Internal error. Now exiting\n");
		printf("String was %s\n", action.toAscii().data());
		exit(-1);
	}

	// Enable or disable the core
	cpu_control->set_enabled(core_selected, enable_cpu);

	// Update the text on the enabled button
	update_enabled_button();
}

/**
 * Play video with video acceleration
 */
void
Theatre::play_accelerated()
{
	gst_player->set_mode(ACCELERATED);
	gst_bus = gst_player->get_bus();
	play_pause();
}

/**
 * Play video without video acceleration
 */
void
Theatre::play_unaccelerated()
{
	gst_player->set_mode(UNACCELERATED);
	gst_bus = gst_player->get_bus();
	play_pause();
}

/**
 * Collect camera input and output it to the screen
 */
void
Theatre::start_camera()
{
	gst_player->set_mode(CAMERA);
	gst_bus = gst_player->get_bus();
	play_pause();
}

/**
 * Load CPU with additional threads
 */
void
Theatre::load_cpu()
{
	bool result;

	// Add extra threads to CPU load
	result = cpu_control->load_cpu();

	// If returned false, we have likely maxed out the number of threads
	if (result == false)
		ui.loadCPUButton->setText(QString::fromUtf8("MAX"));
	else
		ui.loadCPUButton->setText(QString::fromUtf8("++"));

	update_thread_count();
}

/**
 * Remove CPU worker threads to return to default, pristine workload
 * for all CPU cores
 */
void
Theatre::unload_cpu()
{
	// Remove all background threads from CPUControl
	cpu_control->unload_cpu();

	// Reset text on load-CPU button
	ui.loadCPUButton->setText(QString::fromUtf8("+"));

	update_thread_count();
}

void
Theatre::update_thread_count()
{
	unsigned int num_threads = cpu_control->get_num_threads();
	const unsigned int MAX_LENGTH = 200;
	char thread_label_text[MAX_LENGTH];
	char load_button_text[MAX_LENGTH];

	sprintf(thread_label_text, "Background thread count: %d\n", num_threads);
	ui.threadLabel->setText(QString::fromUtf8(thread_label_text));

	/*
	sprintf(load_button_text,
			"Add CPU load (more threads)\nCurrent count: %d",
			num_threads);
	ui.loadCPULabel->setText(QString::fromUtf8(load_button_text));
	*/

}
